
#include "x.h"
#include "xt.h"
#include "xpex.h"
#include <math.h>
#include <stdlib.h>
#include <string.h>
#include "book_utils.h"
#include "externs.h"
#include "defs.h"
#include "globals.h"
#include "color.h"

Display *dpy;
Window   window;

XtInputId specInput, matInput;

PEXRenderer           renderer;
PEXRendererAttributes rattrs;

PEXStructure top_struct;
Atom         delete_window_atom;
Widget       toplevel, canvas;

/* Callback function to quit the program. */
void
quit( w, client_data, call_data )
Widget     w;
caddr_t    client_data;
XtPointer *call_data;
{
#if 0
    XClientMessageEvent        *cm = &call_data->event->xclient;

    if ( cm->data.l[0] == delete_window_atom ) {
        XtDestroyWidget( toplevel );
        exit( 0 );
    }
#endif
}


/* Callback for Expose events. */
void
expose( w, client_data, event )
Widget    w;
XtPointer    client_data;
XEvent    *event;
{
    /* Redraw when last Expose event received. */
    if ( event->xexpose.count == 0 )
        redraw();
}


void
set_refl_props_color( cbs )
char *cbs;
{
    char type[3];
    float val1, val2, val3;
    String string = (String) malloc(26);
    extern Widget reflectColorLabel;
    extern PEXReflectionAttributes refl_props;

    XFlush( dpy );

    PEXSetEditingMode( dpy, top_struct, PEXStructureReplace );
    PEXSetElementPtr( dpy, top_struct, PEXBeginning, 0 );
    PEXSetElementPtrAtLabel( dpy, top_struct, REFL_PROPS_LABEL, 1 );

    sscanf( cbs, "%s %f %f %f\n", &type, &val1, &val2, &val3 );
    if( !strcmp( type, "RGB" ) ) {
        refl_props.specular_color.type = PEXColorTypeRGB;
        refl_props.specular_color.value.rgb.red = val1;
        refl_props.specular_color.value.rgb.green = val2;
        refl_props.specular_color.value.rgb.blue = val3;
        sprintf( string, "R: %.2f G: %.2f B: %.2f", val1, val2, val3 );
        XtVaSetValues( reflectColorLabel, XtNlabel, string, NULL );
        PEXSetReflectionAttributes( dpy, top_struct, PEXOCStore, &refl_props );
    }
    else if( !strcmp( type, "HSV" ) ) {
        refl_props.specular_color.type = PEXColorTypeHSV;
        refl_props.specular_color.value.hsv.hue = val1;
        refl_props.specular_color.value.hsv.saturation = val2;
        refl_props.specular_color.value.hsv.value = val3;
        sprintf( string, "H: %.2f S: %.2f V: %.2f", val1, val2, val3 );
        XtVaSetValues( reflectColorLabel, XtNlabel, string, NULL );
        if( !colorTypeHSV ) {
            HSV hsv_values;
            RGB rgb_values;
            PEXReflectionAttributes temp;

            hsv_values.h = val1;
            hsv_values.s = val2;
            hsv_values.v = val3;
            rgb_values = HSVToRGB( hsv_values );
            temp.ambient = refl_props.ambient;
            temp.diffuse = refl_props.diffuse;
            temp.specular = refl_props.specular;
            temp.specular_conc = refl_props.specular_conc;
            temp.transmission = refl_props.transmission;
            temp.specular_color.type = PEXColorTypeRGB;
            temp.specular_color.value.rgb.red =
                (float)rgb_values.r / 65535.0;
            temp.specular_color.value.rgb.green =
                (float)rgb_values.g / 65535.0;
            temp.specular_color.value.rgb.blue =
                (float)rgb_values.b / 65535.0;
            PEXSetReflectionAttributes( dpy, top_struct, PEXOCStore, &temp );
        }
        else
            PEXSetReflectionAttributes( dpy, top_struct, PEXOCStore, &refl_props );
    }
    else if( !strcmp( type, "HLS" ) ) {
        refl_props.specular_color.type = PEXColorTypeHLS;
        refl_props.specular_color.value.hls.hue = val1;
        refl_props.specular_color.value.hls.lightness = val2;
        refl_props.specular_color.value.hls.saturation = val3;
        sprintf( string, "H: %.2f L: %.2f S: %.2f", val1, val2, val3 );
        XtVaSetValues( reflectColorLabel, XtNlabel, string, NULL );
        if( !colorTypeHLS ) {
            HLS hls_values;
            RGB rgb_values;
            PEXReflectionAttributes temp;

            hls_values.h = val1;
            hls_values.l = val2;
            hls_values.s = val3;
            rgb_values = HLSToRGB( hls_values );
            temp.ambient = refl_props.ambient;
            temp.diffuse = refl_props.diffuse;
            temp.specular = refl_props.specular;
            temp.specular_conc = refl_props.specular_conc;
            temp.transmission = refl_props.transmission;
            temp.specular_color.type = PEXColorTypeRGB;
            temp.specular_color.value.rgb.red =
                (float)rgb_values.r / 65535.0;
            temp.specular_color.value.rgb.green =
                (float)rgb_values.g / 65535.0;
            temp.specular_color.value.rgb.blue =
                (float)rgb_values.b / 65535.0;
            PEXSetReflectionAttributes( dpy, top_struct, PEXOCStore, &temp );
        }
        else
            PEXSetReflectionAttributes( dpy, top_struct, PEXOCStore, &refl_props );
    }
    else if( !strcmp( type, "CIE" ) ) {
        refl_props.specular_color.type = PEXColorTypeCIE;
        refl_props.specular_color.value.cie.x = val1;
        refl_props.specular_color.value.cie.y = val2;
        refl_props.specular_color.value.cie.z = val3;
        sprintf( string, "X: %.2f Y: %.2f Z: %.2f", val1, val2, val3 );
        XtVaSetValues( reflectColorLabel, XtNlabel, string, NULL );
        if( !colorTypeCIE ) {
            CIE cie_values;
            RGB rgb_values;
            PEXReflectionAttributes temp;

            cie_values.X = val1;
            cie_values.Y = val2;
            cie_values.Z = val3;
            rgb_values = CIEToRGB( cie_values );
            temp.ambient = refl_props.ambient;
            temp.diffuse = refl_props.diffuse;
            temp.specular = refl_props.specular;
            temp.specular_conc = refl_props.specular_conc;
            temp.transmission = refl_props.transmission;
            temp.specular_color.type = PEXColorTypeRGB;
            temp.specular_color.value.rgb.red =
                (float)rgb_values.r / 256.0;
            temp.specular_color.value.rgb.green =
                (float)rgb_values.g / 256.0;
            temp.specular_color.value.rgb.blue =
                (float)rgb_values.b / 256.0;
            PEXSetReflectionAttributes( dpy, top_struct, PEXOCStore, &temp );
        }
        else
            PEXSetReflectionAttributes( dpy, top_struct, PEXOCStore, &refl_props );
    }

    XtRemoveInput( specInput );
    redraw();
}


void
set_material_color( cbs )
char *cbs;
{
    char type[3];
    float val1, val2, val3;
    String string = (String) malloc(26);
    extern Widget materialColorLabel;
    extern PEXColor colorMaterial;
    extern PEXColorType materialColorType;

    XFlush( dpy );

    PEXSetEditingMode( dpy, top_struct, PEXStructureReplace );
    PEXSetElementPtr( dpy, top_struct, PEXBeginning, 0 );
    PEXSetElementPtrAtLabel( dpy, top_struct, MATERIAL_COLOR_LABEL, 1 );

    sscanf( cbs, "%s %f %f %f\n", &type, &val1, &val2, &val3 );
    if( !strcmp( type, "RGB" ) ) {
         colorMaterial.rgb.red = val1;
         colorMaterial.rgb.green = val2;
         colorMaterial.rgb.blue = val3;
         sprintf( string, "R: %.2f G: %.2f B: %.2f", val1, val2, val3 );
         XtVaSetValues( materialColorLabel, XtNlabel, string, NULL );
         materialColorType = PEXColorTypeRGB;
         PEXSetSurfaceColor( dpy, top_struct, PEXOCStore, PEXColorTypeRGB, &colorMaterial );
    }
    else if( !strcmp( type, "HSV" ) ) {
         colorMaterial.hsv.hue = val1;
         colorMaterial.hsv.saturation = val2;
         colorMaterial.hsv.value = val3;
         sprintf( string, "H: %.2f S: %.2f V: %.2f", val1, val2, val3 );
         XtVaSetValues( materialColorLabel, XtNlabel, string, NULL );
         materialColorType = PEXColorTypeHSV;
         if( !colorTypeHSV ) {
             HSV hsv_values;
             RGB rgb_values;
             PEXColor temp;
             
             hsv_values.h = val1;
             hsv_values.s = val2;
             hsv_values.v = val3;
             rgb_values = HSVToRGB( hsv_values );
             temp.rgb.red = (float)rgb_values.r / 65535.0;
             temp.rgb.green = (float)rgb_values.g / 65535.0;
             temp.rgb.blue = (float)rgb_values.b / 65535.0;
             PEXSetSurfaceColor( dpy, top_struct, PEXOCStore, PEXColorTypeRGB, &temp );
         }
         else
             PEXSetSurfaceColor( dpy, top_struct, PEXOCStore, PEXColorTypeHSV, &colorMaterial );
    }
    else if( !strcmp( type, "HLS" ) ) {
         colorMaterial.hls.hue = val1;
         colorMaterial.hls.lightness = val2;
         colorMaterial.hls.saturation = val3;
         sprintf( string, "H: %.2f L: %.2f S: %.2f", val1, val2, val3 );
         XtVaSetValues( materialColorLabel, XtNlabel, string, NULL );
         materialColorType = PEXColorTypeHLS;
         if( !colorTypeHLS ) {
             HLS hls_values;
             RGB rgb_values;
             PEXColor temp;
             
             hls_values.h = val1;
             hls_values.l = val2;
             hls_values.s = val3;
             rgb_values = HLSToRGB( hls_values );
             temp.rgb.red = (float)rgb_values.r / 65535.0;
             temp.rgb.green = (float)rgb_values.g / 65535.0;
             temp.rgb.blue = (float)rgb_values.b / 65535.0;
             PEXSetSurfaceColor( dpy, top_struct, PEXOCStore, PEXColorTypeRGB, &temp );
         }
         else
             PEXSetSurfaceColor( dpy, top_struct, PEXOCStore, PEXColorTypeHLS, &colorMaterial );
    }
    else if( !strcmp( type, "CIE" ) ) {
         colorMaterial.cie.x = val1;
         colorMaterial.cie.y = val2;
         colorMaterial.cie.z = val3;
         sprintf( string, "X: %.2f Y: %.2f Z: %.2f", val1, val2, val3 );
         XtVaSetValues( materialColorLabel, XtNlabel, string, NULL );
         materialColorType = PEXColorTypeCIE;
         if( !colorTypeCIE ) {
             CIE cie_values;
             RGB rgb_values;
             PEXColor temp;

             cie_values.X = val1;
             cie_values.Y = val2;
             cie_values.Z = val3;
             rgb_values = CIEToRGB( cie_values );
             temp.rgb.red = (float)rgb_values.r / 256.0;
             temp.rgb.green = (float)rgb_values.g / 256.0;
             temp.rgb.blue = (float)rgb_values.b / 256.0;
             PEXSetSurfaceColor( dpy, top_struct, PEXOCStore,
                 PEXColorTypeRGB, &temp );
         }
         else
             PEXSetSurfaceColor( dpy, top_struct, PEXOCStore,
                 PEXColorTypeCIE, &colorMaterial );
    }

    XtRemoveInput( matInput );
    redraw();
}


void
set_refl_props( widget, parameter, cbs )
Widget    widget;
caddr_t   parameter;
XtPointer cbs;
{
    extern Widget reflectAmbientLabel, reflectDiffuseLabel,
                  reflectTransmissionLabel, reflectSpecularLabel,
                  reflectSpecularConcLabel;
    String string;
    float per = *( float* )cbs;
    float top = ( float )per;
    extern PEXReflectionAttributes    refl_props;

    string = (String) malloc(10);
    sprintf( string, "%.2f", top );

    switch ( (int)parameter ) {
    case AMBIENT_COEF:
        refl_props.ambient = top;
        XtVaSetValues( reflectAmbientLabel, XtNlabel, string, NULL );
        break;
    case DIFFUSE_COEF:
        refl_props.diffuse = top;
        XtVaSetValues( reflectDiffuseLabel, XtNlabel, string, NULL );
        break;
    case TRANSMISSION_COEF:
        refl_props.transmission = top;
        XtVaSetValues( reflectTransmissionLabel, XtNlabel, string, NULL );
        break;
    case SPECULAR_COEF:
        refl_props.specular = top;
        XtVaSetValues( reflectSpecularLabel, XtNlabel, string, NULL );
        break;
    case SPECULAR_EXP:
        refl_props.specular_conc = 100 * top;
        sprintf( string, "%.0f", 100 * top );
        XtVaSetValues( reflectSpecularConcLabel, XtNlabel, string, NULL );
        break;
    }

    PEXSetEditingMode( dpy, top_struct, PEXStructureReplace );
    PEXSetElementPtr( dpy, top_struct, PEXBeginning, 0 );
    PEXSetElementPtrAtLabel( dpy, top_struct, REFL_PROPS_LABEL, 1 );
    PEXSetReflectionAttributes( dpy, top_struct, PEXOCStore, &refl_props );
    PEXSetElementPtr( dpy, top_struct, PEXCurrent, 1 );
    PEXSetBFReflectionAttributes( dpy, top_struct, PEXOCStore, &refl_props );
    redraw();
}


void editSpecColor( widget, call_data, client_data )
Widget    widget;
XtPointer call_data, client_data;
{
    extern XtAppContext app;
    extern char *getenv();
    extern PEXReflectionAttributes refl_props;
    extern void get_specular_reply();

    if( getenv( "COLOR_EDITOR" ) ) {
        char *program, *args;

        specInput = XtAppAddInput( app, fileno(stdin), (XtPointer)XtInputReadMask, get_specular_reply, NULL );

        program = (char *) malloc( strlen(getenv( "COLOR_EDITOR" )) + 25 );
        strcpy( program, getenv( "COLOR_EDITOR" ) );
        args = (char *) malloc(25);
        sprintf( args, " RGB %.2f %.2f %.2f", refl_props.specular_color.value.rgb.red, refl_props.specular_color.value.rgb.green, refl_props.specular_color.value.rgb.blue );
        strcat( program, args );
        xs_talkto( program );
    }
    else {
        fprintf( stderr, "Environment variable COLOR_EDITOR, undefined!!!\n\n");
        XtDestroyApplicationContext( app );
        exit( 0 );
    }
}


void editMaterialColor( widget, call_data, client_data )
Widget    widget;
XtPointer call_data, client_data;
{
    extern XtAppContext app;
    extern char *getenv();
    extern PEXColor colorMaterial;
    extern void get_material_reply();

    if( getenv( "COLOR_EDITOR" ) ) {
        char *program, *args;

        matInput = XtAppAddInput( app, fileno(stdin), (XtPointer)XtInputReadMask, get_material_reply, NULL );

        program = (char *) malloc( strlen(getenv( "COLOR_EDITOR" )) + 25 );
        strcpy( program, getenv( "COLOR_EDITOR" ) );
        args = (char *) malloc(25);
        sprintf( args, " RGB %.2f %.2f %.2f", colorMaterial.rgb.red, colorMaterial.rgb.green, colorMaterial.rgb.blue );
        strcat( program, args );
        xs_talkto( program );
    }
    else {
        fprintf( stderr, "Environment variable COLOR_EDITOR, undefined!!!\n\n");
        XtDestroyApplicationContext( app );
        exit( 0 );
    }
}


void get_specular_reply(client_data, fid, id)
XtPointer    client_data;
int         *fid;
XtInputId   *id;
{
    char       buf[BUFSIZ];
    int        nbytes, i;

    /* Get all pending input */
    nbytes = read(*fid, buf, BUFSIZ);

    /*
     * Null terminate the string at the first newline,
     * or at the end of the bytes read.
     */
    if (nbytes && buf[0] != '\n') {
        for( i= 0; i< nbytes; i++ )
            if(buf[i] == '\n')
                buf[i] = '\0';
        buf[nbytes] = '\0';
        set_refl_props_color( buf );
    }
}


void get_material_reply(client_data, fid, id)
XtPointer    client_data;
int         *fid;
XtInputId   *id;
{
    char       buf[BUFSIZ];
    int        nbytes, i;

    /* Get all pending input */
    nbytes = read(*fid, buf, BUFSIZ);

    /*
     * Null terminate the string at the first newline,
     * or at the end of the bytes read.
     */
    if (nbytes && buf[0] != '\n') {
        for( i= 0; i< nbytes; i++ )
            if(buf[i] == '\n')
                buf[i] = '\0';
        buf[nbytes] = '\0';
        set_material_color( buf );
    }
}
